/* bsd source causes __FAVOR_BSD to be defined in tcp.h */
#define _BSD_SOURCE
/* include the packet capture library */
#include<pcap.h>
/* include the command line option handling library */
#include<getopt.h>
/* the basics */
#include<stdlib.h>
#ifdef HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif
#include <sys/stat.h>
#include <unistd.h>
#include <string.h>
#include <assert.h>
/* ip, tcp and ethernet headers */
#include <netinet/ip.h>
#include <netinet/tcp.h>
#include <netinet/if_ether.h>

/* my hashtable functions */
#include "hashtable.h"

#define MIN(a,b) (((a)<(b)) ? (a) : (b))
#define MAX(a,b) (((a)>(b)) ? (a) : (b))

/* add file handles here for additional plots */
static FILE *f_seqplot;
static FILE *f_ackplot;

/* stuff you'll use to handle round trip time estimation */
/* this is the structure entered into the hashtable on transmission */
struct packetSendTime {
  u_int32_t sequence_number;
  struct timeval ts;
};
/* hash table consisting of the above */
static hashtable transmission_times; 
/* hash table helper functions - hash and isequal */
static unsigned int hash_seq(const void *k);
static boolean isequal_seq(const void *k1, const void *k2);

/* what is bigger - smaller expressed in double precision
   floating point seconds? */
static double delta_time(const struct timeval *bigger, 
                  const struct timeval *smaller);

/* what is the time since the first packet we send to this
   function?  used to make the time axis start at zero for
   all graphs. */
static double time_since_start(const struct timeval *t);

/* utility function used to dump packet data to stdout in
   hexadecimal with some annotation */
static void spewpack(const u_char *wholePacket, 
              unsigned short length, 
              unsigned int strip_length);

/* modify these functions to add additional files for time value pairs */
static void open_output_files ( const char * basename ) {
  char filename[255];

  strcpy(filename, basename);
  strcat(filename, ".seq");
  f_seqplot = fopen(filename, "w");
  assert(f_seqplot != NULL);

  strcpy(filename, basename);
  strcat(filename, ".ack");
  f_ackplot = fopen(filename, "w");
  assert(f_ackplot != NULL);





}

static void close_output_files ( void ) {
  (void)fclose(f_seqplot);
  (void)fclose(f_ackplot);
}


/* called for each packet in the trace. dummy is not needed here,
   header is defined in /usr/include/pcap.h (or on stupid redhat
   machines like sumatra /usr/include/pcap/pcap.h) */
static void handle_packet(u_char *dummy, 
                   const struct pcap_pkthdr *header, 
                   const u_char *pktdata) {
  /* skip ethernet header to find the ip header */
  struct ip *ip   = (struct ip *)(pktdata + ETHER_HDR_LEN);  
  /* the tcp header */
  struct tcphdr *tcp;  
  /* time since beginning of trace */
  double current_time;   
  /* sequence number at b.of t.*/
  static u_int32_t start_sequence_number; 


  /* verify that we found the ip header.  This might fail if
     the trace was taken of a connection on a media type
     other than ethernet.  The added complexity of handling
     different media types didn't seem necessary for this
     assignment */
  if(ip->ip_v != 4) {
    printf("catastrophic error: not aligned\n");
    spewpack(pktdata, header->caplen, ETHER_HDR_LEN);
    exit(EXIT_FAILURE);
  } 
  
  /* find the tcp header as an offset off the ip header */
  tcp = (struct tcphdr *)((char *)ip + 4*ip->ip_hl);

  /* what time is it, relative to the beginning of the trace */
	/* the type cast is from a strange naming in redhat's 
	 * bundled pcap library. */
  current_time = time_since_start((const struct timeval *)&header->ts);
  
  if(ntohs(tcp->th_sport) == 80) {
    /* traffic from the web server */
    struct packetSendTime *trans = malloc(sizeof(struct packetSendTime));

    /* the length of the data stream is inferred from the length
       of the packet and the header stuff.  This isn't right in
       general, considering the possiblility of urgent data. */
    unsigned int data_length = 
      ntohs(ip->ip_len)    /* length of the datagram */
      - (ip->ip_hl * 4)    /* minus ip header length */
      - ((int)tcp->th_off * 4); /* minus tcp header length */
    
    assert(trans != NULL);

    /* set the starting sequence number for this connection,
       if not yet set.  has a 1 in 4 billion chance of being
       off by one. */
    if(start_sequence_number == 0) 
      start_sequence_number = ntohl(tcp->th_seq);

    /* print time, (end) sequence number pair to f_seqplot */
    fprintf(f_seqplot, "%f %u\n", current_time, 
            ntohl(tcp->th_seq) - 
            start_sequence_number + 
            data_length );


    trans->sequence_number = ntohl(tcp->th_seq) + data_length;
    trans->ts.tv_sec           = header->ts.tv_sec;
    trans->ts.tv_usec          = header->ts.tv_usec;
    ht_insert(transmission_times, trans, trans);

  } else if(ntohs(tcp->th_dport) == 80) {
    /* traffic to the web server */

    /* the first return path packet (syn from the client)
       doesn't have the ack flag set, so the th_ack field is
       undefined. */
    if((tcp->th_flags & TH_ACK) != 0) {
      const struct packetSendTime *trans;
      u_int32_t lookup = ntohl(tcp->th_ack);
      /* print time, ack sequence number pair to f_ackplot */
      fprintf(f_ackplot, "%f %u\n", current_time, 
              ntohl(tcp->th_ack) - start_sequence_number);



      trans = ht_lookup(transmission_times, &lookup);
    }
  } else {
    /* other traffic, what is it doing here? */
    printf("non-web traffic: %f (%u:%u) %u\n", current_time,
           ntohs(tcp->th_sport), ntohs(tcp->th_dport),
           ntohl(tcp->th_seq));
  }

}


static pcap_t *open_pcap(const char *filename, 
                  const char *cmd_pcap_filter);
static void print_usage(void);

/* handle command line arguments, open files, then loop over
   each packet calling handle_packet() */
int main(int argc, char *argv[]) {
  static const struct option getopt_long_options[] = {
    /* tcpdump formatted input file: */
    { "readfile",   required_argument, NULL, (int)'r' }, 
    /* pcap-style filter; may need to be quoted */
    { "filter",     required_argument, NULL, (int)'f' }, 
    /* basename for the output coordinate pairs to be graphed,
       if different from readfile */
    { "outputbase", required_argument, NULL, (int)'o' }, 
    { "help", no_argument, NULL, (int)'h' }, 
    { NULL,0,NULL,0 } /* null termination of the array */
  };
  int ci, option_idx=0;
  const char *cmd_readfile = NULL;
  const char *cmd_pcap_filter = "ip";
  const char *cmd_basename = NULL;
  pcap_t *pcap_trace;
  while((ci = getopt_long(argc,argv, "r:f:o:h", getopt_long_options, 
                          &option_idx)) != -1) {
    switch ((char)ci) {
    case 'r':
      cmd_readfile = strdup(optarg);
      break;
    case 'f':
      cmd_pcap_filter = strdup(optarg);
      break;
    case 'o':
      cmd_basename = strdup(optarg);
      break;
    case 'h':
      print_usage(); 
      exit(EXIT_SUCCESS);
    default:
      fprintf(stderr, "unexpected error parsing command line arguments\n");
      print_usage(); 
      exit(EXIT_FAILURE);
    }
  }
  if(cmd_readfile == NULL) {
    print_usage(); 
    exit(EXIT_FAILURE);
  }
  if(cmd_basename == NULL) {
    int i;
    cmd_basename = strdup(cmd_readfile);
    for(i=0; cmd_basename[i]!='\0';i++) {
      if(cmd_basename[i]=='.') ((char *)cmd_basename)[i] = '\0'; 
    }
  }

  pcap_trace = open_pcap(cmd_readfile, cmd_pcap_filter);

  open_output_files ( cmd_basename );

  transmission_times = ht_new(1024, hash_seq, isequal_seq, NULL, free);

  (void)pcap_loop(pcap_trace, 0, handle_packet, NULL);
  
  close_output_files ( );

  ht_delete(transmission_times);

  free((char *)cmd_basename);
  free((char *)cmd_readfile);
  exit(EXIT_SUCCESS);
}

/* utility functions you may or may not need to use, and
   probably shouldn't modify. */

/* print the bytes in the packet in hexadecimal, annotaing
   the end of the link-level header (#), and the presumed
   end of the ip header (@) */
static void spewpack(const u_char *wholePacket, 
              unsigned short length, 
              unsigned int strip_length) {
  unsigned int i;
  for (i=0; i<MIN(strip_length+40, length); i++) {
    if(i==strip_length)    fprintf(stderr, "#");
    if(i==strip_length+20) fprintf(stderr, "@");
    fprintf(stderr, "%02x",(unsigned int)wholePacket[i]);
    if(i%4==3)             fprintf(stderr, " ");
  }
  fprintf(stderr, "\n");
}

static double delta_time(const struct timeval *bigger, 
                const struct timeval *smaller) {
  return(  (double)(bigger->tv_sec-smaller->tv_sec) 
         + ((double) (bigger->tv_usec - smaller->tv_usec)/1000000.0));
}
/* calculate the difference in time since this function was
 called first. start_time is stored across invocations of
 this function (static), and starts out being 0.  Return the
 time difference as a double (float) for easy output. */
static double time_since_start(const struct timeval *t) {
  static struct timeval start_time;
  if(start_time.tv_sec == 0 && start_time.tv_usec == 0) {
    start_time.tv_sec = t->tv_sec;
    start_time.tv_usec = t->tv_usec;
    return(0.0);
  }
  return(delta_time(t, &start_time));
}

/* print an arguably helpful message about how to run the
   program. */
static void print_usage(void) {
  printf("Usage: sequencer {arguments} where {arguments} are:\n"
         "       --readfile tcpdumpfile  (required)\n"
         "             specifies which file to open for analysis\n"
         "       --filter \"pcapfilter string\"  (optional, default is ip)\n"
         "             specifies which packets to examine, see man tcpdump\n"
         "       --outputbase prefix  (optional, default is tcpdumpfile)\n"
         "             what prefix to apply to filenames for graph output\n"
         " -r, -f, and -o are legal abbreviations for the long forms\n"
         " eg.  ./sequencer -r mytrace -f \"ip src 10.239.2.1\" -o home\n");
}

/* global buffer to hold errors from pcap activities */
static char errbuf[PCAP_ERRBUF_SIZE];
/* set up the pcap file for reading.  YOU SHOULD NOT NEED TO
   MODIFY THIS FUNCTION */
static pcap_t *open_pcap(const char *filename, 
                  const char *cmd_pcap_filter) {
  struct bpf_program ip_filter[2];
  int zSuccess;
  pcap_t *pd; /* return value */ 
  struct stat s;
  errbuf[0]='\0';

  pd=pcap_open_offline(filename, errbuf);
  if(pd == NULL) {
    fprintf(stderr, "Error opening file \"%s\": %s.\n", filename, errbuf); 
    exit(EXIT_FAILURE); 
  }
  if(stat(filename,&s)== 0) {
    unsigned long file_length =  (unsigned long)s.st_size;
    printf("opening file \"%s\" of %lu bytes.\n", filename, file_length); 
  }

  /* int pcap_compile(pcap_t *p, struct bpf_program *fp,
           char *str, int optimize, bpf_u_int32 netmask) */
  zSuccess=pcap_compile(pd, &ip_filter[0], (char *)cmd_pcap_filter, 1, 0);
  if(zSuccess != 0) {
    pcap_perror(pd,(char *)"your filter didn't compile\n");
    fprintf(stderr, "The offending filter was: %s\n", cmd_pcap_filter);
    exit(EXIT_FAILURE);
  }
    
  /* int pcap_setfilter(pcap_t *p, struct bpf_program *fp) */
  zSuccess=pcap_setfilter(pd, ip_filter);
  if(zSuccess != 0) {
    pcap_perror(pd,(char *)"no filter\n");
    exit(EXIT_FAILURE);
  }
  return(pd);
}

/* helper functions for the hash table, 
   hashes and compares (indirect) integers */
static unsigned int hash_seq(const void *k) {
  const u_int32_t *key = (const u_int32_t *) k;
  return(*key);
}
static boolean isequal_seq(const void *k1, const void *k2) {
  const u_int32_t *key1 = (const u_int32_t *)k1;
  const u_int32_t *key2 = (const u_int32_t *)k2;
  return(*key1 == *key2);
}
